from __future__ import annotations

from collections.abc import Container, Iterable
from typing import TYPE_CHECKING

from liblaf.grapes import magic
from liblaf.grapes.logging import autolog

if TYPE_CHECKING:
    from _typeshed import SupportsGetItem


_DEPRECATED_MESSAGE = "'%s' is deprecated. Please use '%s' instead."


def contains[T](
    obj: Container[T],
    key: T,
    deprecated_keys: Iterable[T] = (),
    *,
    msg: str = _DEPRECATED_MESSAGE,
) -> bool:
    __warnings_hide = True
    if key in obj:
        return True
    for deprecated_key in deprecated_keys:
        if deprecated_key in obj:
            stacklevel: int
            _, stacklevel = magic.get_frame_with_stacklevel(
                hidden=magic.hidden_from_warnings
            )
            autolog.warning(msg, deprecated_key, key, stacklevel=stacklevel)
            return True
    return False


def getitem[KT, VT](
    obj: SupportsGetItem[KT, VT],
    key: KT,
    deprecated_keys: Iterable[KT] = (),
    *,
    msg: str = _DEPRECATED_MESSAGE,
) -> VT:
    __warnings_hide = True
    try:
        return obj[key]
    except KeyError:
        pass
    for deprecated_key in deprecated_keys:
        try:
            value: VT = obj[deprecated_key]
        except KeyError:
            continue
        stacklevel: int
        _, stacklevel = magic.get_frame_with_stacklevel(
            hidden=magic.hidden_from_warnings
        )
        autolog.warning(msg, deprecated_key, key, stacklevel=stacklevel)
        return value
    raise KeyError(key)



import inspect
import types
from collections.abc import Callable, Iterable, Mapping

from liblaf.grapes._config import config

from ._release_type import is_pre_release

_HIDDEN_FROM_LOGGING_NAMES: tuple[str, ...] = ("rich.progress.",)


def hidden_from_logging(frame: types.FrameType | None) -> bool:
    if frame is None:
        return False
    if _getitem(frame.f_locals, ("__tracebackhide__", "__logging_hide"), default=False):
        return True
    name: str = frame.f_globals.get("__name__", "")
    return f"{name}.".startswith(_HIDDEN_FROM_LOGGING_NAMES)


def hidden_from_traceback(
    frame: types.FrameType | None, *, hide_stable_release: bool | None = None
) -> bool:
    if frame is None:
        return False
    if _getitem(frame.f_locals, ("__tracebackhide__",), default=False):
        return True
    if hide_stable_release is None:
        hide_stable_release = config.traceback.hide_stable_release.get()
    if hide_stable_release:
        name: str | None = frame.f_globals.get("__name__")
        if not is_pre_release(frame.f_code.co_filename, name):
            return True
    return False


def hidden_from_warnings(
    frame: types.FrameType | None, *, hide_stable_release: bool | None = None
) -> bool:
    if frame is None:
        return False
    if _getitem(
        frame.f_locals, ("__tracebackhide__", "__warnings_hide"), default=False
    ):
        return True
    if hide_stable_release is None:
        hide_stable_release = config.traceback.hide_stable_release.get()
    if hide_stable_release:
        name: str | None = frame.f_globals.get("__name__")
        if not is_pre_release(frame.f_code.co_filename, name):
            return True
    return False


def get_frame(
    depth: int = 1, hidden: Callable[[types.FrameType], bool] | None = None
) -> types.FrameType | None:
    frame: types.FrameType | None = inspect.currentframe()
    while frame is not None and depth > 0:
        frame = frame.f_back
        depth -= 1
        if hidden is not None:
            while frame is not None and hidden(frame):
                frame = frame.f_back
    return frame


def get_frame_with_stacklevel(
    depth: int = 1, hidden: Callable[[types.FrameType], bool] | None = None
) -> tuple[types.FrameType | None, int]:
    frame: types.FrameType | None = inspect.currentframe()
    stacklevel: int = 0
    while frame is not None and depth > 0:
        frame = frame.f_back
        depth -= 1
        stacklevel += 1
        if hidden is not None:
            while frame is not None and hidden(frame):
                frame = frame.f_back
                stacklevel += 1
    return frame, stacklevel


def _getitem[KT, VT](obj: Mapping[KT, VT], keys: Iterable[KT], default: VT) -> VT:
    for key in keys:
        try:
            return obj[key]
        except KeyError:
            continue
    return default



from collections.abc import Generator
from typing import Any

import fieldz
import wrapt
from rich.repr import RichReprResult

from liblaf.grapes.sentinel import MISSING
from liblaf.grapes.typing import is_array
from liblaf.grapes.wadler_lindig import pformat


class ReprWrapper(wrapt.ObjectProxy):
    def __repr__(self) -> str:
        return pformat(self.__wrapped__)


def rich_repr_fieldz(obj: object) -> RichReprResult:
    for name, value, default in _iter_fieldz(obj):
        if default is MISSING:
            yield name, wraps_repr(value)
        else:
            yield name, wraps_repr(value), wraps_repr(default)


def wraps_repr[T](obj: T) -> T:
    if is_array(obj):
        return ReprWrapper(obj)  # pyright: ignore[reportReturnType]
    return obj


def _iter_fieldz(obj: object) -> Generator[tuple[str, Any, Any]]:
    for field in fieldz.fields(obj):
        if not field.repr:
            continue
        value: Any = getattr(obj, field.name, MISSING)
        # rich.repr uses `if default == value:` but does not protect against
        # exceptions. Some types (e.g. NumPy arrays) raise on equality/truth
        # checks (ambiguous truth value). Do the comparison here inside a
        # try/except so we can catch and handle those errors safely.
        try:
            if value == field.default:
                yield field.name, value, field.default
            else:
                yield field.name, value, MISSING
        except Exception:  # noqa: BLE001
            yield field.name, value, MISSING



from collections.abc import Callable, Generator, Iterable
from typing import Any

import attrs
import wrapt


@attrs.define
class CacheIter[T]:
    _iter: Iterable[T]
    _cache: list[T] = attrs.field(repr=False, init=False, factory=list)

    def __iter__(self) -> Generator[T]:
        yield from self._cache
        for item in self._iter:
            self._cache.append(item)
            yield item


def cache_iter[C: Callable[..., Iterable]](func: C) -> C:
    @wrapt.decorator
    def wrapper[T](
        wrapped: Callable[..., Iterable[T]],
        _instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Iterable[T]:
        return CacheIter(wrapped(*args, **kwargs))

    return wrapper(func)  # pyright: ignore[reportReturnType]



import functools
import logging
from collections.abc import Iterable
from typing import Any

import attrs
import limits


@attrs.define
class LimitsFilter:
    limiter: limits.strategies.RateLimiter = attrs.field(
        factory=lambda: limits.strategies.FixedWindowRateLimiter(
            limits.storage.MemoryStorage()
        )
    )

    def filter(self, record: logging.LogRecord) -> bool:
        args: Any = getattr(record, "limits", None)
        if args is None:
            return True
        hit_args: _HitArgs = _parse_args(args, record)
        return self.limiter.hit(
            hit_args.item, *hit_args.identifiers, cost=hit_args.cost
        )


@attrs.define
class _HitArgs:
    item: limits.RateLimitItem
    identifiers: Iterable[str]
    cost: int = 1


def _default_identifier(record: logging.LogRecord) -> Iterable[str]:
    return record.pathname, record.funcName, str(record.lineno)


@functools.singledispatch
def _parse_args(args: Any, _record: logging.LogRecord) -> _HitArgs:
    raise ValueError(args)


@_parse_args.register(str)
def _parse_args_str(args: str, record: logging.LogRecord) -> _HitArgs:
    return _HitArgs(item=limits.parse(args), identifiers=_default_identifier(record))


@_parse_args.register(limits.RateLimitItem)
def _parse_args_item(args: limits.RateLimitItem, record: logging.LogRecord) -> _HitArgs:
    return _HitArgs(item=args, identifiers=_default_identifier(record))
